;; Generated package description from ggtags.el  -*- no-byte-compile: t -*-
(define-package "ggtags" "0.9.0.0.20201119.3758" "emacs frontend to GNU Global source code tagging system" '((emacs "25")) :keywords '("tools" "convenience") :authors '(("Leo Liu" . "sdl.web@gmail.com")) :maintainer '("Leo Liu" . "sdl.web@gmail.com") :url "https://github.com/leoliu/ggtags")
